import os
import platform
import logging
import shutil
import re
import sys
import time
from threading import Thread

if platform.system() == 'Windows':
    import winreg
    import psutil
    import easygui

APP_TITLE = "PHISHERMON"

# Logging setup
log = logging.getLogger('phishermon')
log.setLevel(logging.DEBUG)
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
log.addHandler(ch)


class MitigateProcess(object):

    def __init__(self):
        pass

    def prompt_kill_process(self, pid, process_name):
        """Asks user if they want terminate the suspicious process.
        
        :param pid: The process ID to be killed.
        :type pid: int
        :param process_name: The name of the process to be killed.
        :type process_name: str

        :returns: True or False depending on whether it was able to kill the process. 
        """

        if psutil.pid_exists(pid) and psutil.Process(pid).name() == process_name:
            self.suspend_process(pid, process_name)

            if self._show_alert_process(pid, process_name):
                self.kill_process(pid, process_name)
                return True
            else:
                self.resume_process(pid,process_name)
                log.info("User did not want to end process: {}-{}".format(process_name, pid))
                return False
        else:
            log.warning("Process: {} with ID: {} does not exist".format(process_name, pid))
            return False

    def _show_alert_process(self, pid, process_name):
        """Helper function displays a prompt asking user if they would like to stop the selected process.
        
        :param pid: The process ID to be killed.
        :type pid: int
        :param process_name: The name of the process to be killed.
        :type process_name: str
        
        :returns: True if the user decided to kill the process, False otherwise.
        """

        msg = ["Phishermon found a suspicious application, would you like to close it?\n",
               "Process Name: {}".format(process_name), 
               "Process ID: {}".format(pid), 
               "{}".format(APP_TITLE)]

        msg = '\n'.join(msg)

        decision = easygui.ccbox(msg)
        return decision

    def kill_process(self, pid, process_name, recursive=False):
        """Helper function that kills process and its children.
        
        :param pid: The process ID to be killed.
        :type pid: int
        :param process_name: The name of the process to be killed.
        :type process_name: str
        :param recursive: If True will kill all child processes of the specified process
        :type recursive: bool
        """

        if pid == os.getpid():
            log.info("I refuse to kill myself")

        if not (type(pid) is int):
            raise RuntimeError("The input must be an integer value representing the Process ID to kill")

        if psutil.pid_exists(pid) and psutil.Process(pid).name() == process_name:
            parent = psutil.Process(pid)

            if recursive:
                for child in parent.children(recursive=True):
                    child.kill()
            parent.kill()

        else:
            log.warning("Process: {} with ID: {} does not exist".format(process_name, pid))

    def suspend_process(self, pid, process_name):
        """Helper function that suspends a process.
        
        :param pid: The process ID to be killed.
        :type pid: int
        :param process_name: The name of the process to be killed.
        :type process_name: str        
        """

        if pid == os.getpid():
            print("I refuse to suspend myself")

        if not isinstance(pid, int):
            raise RuntimeError("The input must be an integer value representing the Process ID to kill")

        if psutil.pid_exists(pid) and psutil.Process(pid).name() == process_name:
            parent = psutil.Process(pid)
            parent.suspend()

        else:
            log.warning("Process: {} with ID: {} does not exist".format(process_name, pid))

    def resume_process(self, pid, process_name):
        """Helper function to resume a suspended process
        
        :param pid: The process ID to be killed.
        :type pid: int
        :param process_name: The name of the process to be killed.
        :type process_name: str  
        """

        if not isinstance(pid, int):
            raise RuntimeError("The input must be an integer value representing the Process ID to kill")

        if psutil.pid_exists(pid) and psutil.Process(pid).name() == process_name:
            parent = psutil.Process(pid)
            parent.resume()

        else:
            log.warning("Process: {} with ID: {} does not exist".format(process_name, pid))

    def get_process_id(self, process_name):
        """Finds process ID by process name, returns None if process not found"""

        for proc in psutil.process_iter():
            if proc.name() == process_name:
                print("Process: " + process_name + ", PID: " + str(proc.pid))
                return proc.pid

        log.warning("No process found with name: " + process_name)
        return None


class MitigateFile(object):

    def __init__(self):
        pass

    def prompt_delete_file(self, path):
        """Asks user if they want delete the suspicious file.
        
        :param path: The path to the file to be deleted
        :type path: str
        
        :returns: True if the file was deleted.
        """

        if self._show_alert_file(path):
            return self.delete_file(path)
        else:
            log.info("User did not want to delete file: {}".format(path))
            return False

    def _show_alert_file(self, path):
        """Helper function displays a prompt asking user if they would like to delete the selected file.
        
        :param path: The path to the file to be deleted
        :type path: str
        """

        text = "Phishermon has found a suspicious file or folder, would you like to delete it?\n\nFile path: {}\n".format(path)

        return easygui.ccbox(text, APP_TITLE)

    def delete_file(self, path):
        """Deletes a file, returns True if file was deleted"""

        if os.path.exists(path):
            if os.path.isdir(path):
                shutil.rmtree(path)
                log.info("Deleted directory at: {}".format(path))
                return True
            else:
                os.remove(path)
                log.info("Deleted file at: {}".format(path))
                return True
        else:
            log.error("The path {} does not exist.".format(path))
            return False


class MitigateRegistry(object):

    def __init__(self):
        self.hive_regex = re.compile(r'(HKEY_[A-Z, _]*)(?:\\)(.*)')

    def prompt_delete_registry_key(self, path, key_to_delete):
        """Asks user if they want delete the suspicious registry key.
        
        :param path: The path to the key to be deleted
        :type path: str
        :param key_to_delete: The key to be deleted
        :type key_to_delete: str
        
        :returns: True if the registry key was deleted.
        """
        key = '{}\\{}'.format(path, key_to_delete)
        if self._show_alert_registry_key(key):
            return self.delete_registry_key(path, key_to_delete)
        else:
            log.info("User did not want to delete registry key: {}".format(key))
            return False

    def _show_alert_registry_key(self, path):
        """Helper function displays a prompt asking user if they would like to delete the selected registry key"""

        text = "Phishermon has found a suspicious registry key, would you like to delete it?\n\nRegistry key path: {}\n".format(path)


        return easygui.ccbox(text, APP_TITLE)

    def delete_registry_key(self, path, key_to_delete):
        """Deletes a registry key tree.
        
        :param path: The path to the key to be deleted
        :type path: str
        :param key_to_delete: The key to be deleted
        :type key_to_delete: str      
        """

        reg, folder = self._parse_reg_path(path)
        log.debug("Inside: {}\\{}".format(folder, key_to_delete))

        try:
            parent_folder = winreg.OpenKey(reg, folder, access=winreg.KEY_ALL_ACCESS | winreg.KEY_WOW64_64KEY)
            to_delete = winreg.OpenKey(reg, folder + '\\' + key_to_delete, access=winreg.KEY_ALL_ACCESS | winreg.KEY_WOW64_64KEY)

            key_info = winreg.QueryInfoKey(to_delete)

            try:
                while True:
                    sub_key = winreg.EnumKey(to_delete, 0)
                    key = '{}\\{}'.format(path, key_to_delete)
                    self.delete_registry_key(key, sub_key)
            except:
                pass

            winreg.DeleteKey(parent_folder, key_to_delete)
            log.info("Deleted registry key: {}\\{}".format(folder, key_to_delete))
        except OSError as e:
            log.error("An error has occured \nError: {0}".format(e))

    def _parse_reg_path(self, path):
        """Parses a full registry key path"""

        match = re.search(self.hive_regex, path)

        if not match:
            raise RuntimeError("Path could not be parsed: " + path)

        hive_name = match.group(1)
        key = match.group(2)

        if hive_name == 'HKEY_CLASSES_ROOT':
            return winreg.HKEY_CLASSES_ROOT, key
        elif hive_name == 'HKEY_CURRENT_USER':
            return winreg.HKEY_CURRENT_USER, key
        elif hive_name == 'HKEY_LOCAL_MACHINE':
            return winreg.HKEY_LOCAL_MACHINE, key
        elif hive_name == 'HKEY_USERS':
            return winreg.HKEY_USERS, key
        elif hive_name == 'HKEY_PERFORMANCE_DATA':
            return winreg.HKEY_PERFORMANCE_DATA, key
        elif hive_name == 'HKEY_CURRENT_CONFIG':
            return winreg.HKEY_CURRENT_CONFIG, key
        else:
            raise RuntimeError("The hive {} does not exist in the Windows registry".format(hive_name))






















import os
import json
import platform
from electus import Electus
from .sysmon_helper import SysmonHelper
from .action import Action
from .utils import format_pretty

class Phishermon(object):

    def __init__(self, library_conf=None, job_conf=None, action_conf=None, evtx_file=None, etw=False, evtx=True, output_file=None, 
                 silent=False, pretty_print=False, continuous=False):

        self.library_conf = library_conf
        self.job_conf = job_conf
        self.action_conf = action_conf
        self.evtx_file = evtx_file
        self.etw = etw
        self.evtx = evtx
        self.output_file = output_file
        self.silent = silent
        self.pretty_print = pretty_print
        self.continuous = continuous

        if self.action_conf:
            if platform.system() == 'Windows' and not self.etw:
                raise ValueError('Actions can only be taken on Windows machines when ETW is True.')
            elif platform.system() != 'Windows':
                raise NotImplementedError('Actions can only be taken on Windows machines.')

        if not self.etw and not self.evtx:
            raise ValueError("No data source specified. At least one of etw or evtx must be True.")

        # Instantiate electus
        self.sig_engine = Electus(self.library_conf, self.job_conf)
        # Instantiate sysmon parser 
        self.sysmon = SysmonHelper(etw=self.etw, evtx=self.evtx, evtx_file=self.evtx_file, continuous=self.continuous)

        # Will return empty list if actions are not defined
        self.actions = self.get_actions()

    @property
    def library_conf(self):
        """Path to an electus library configuration file."""
        return self._library_conf 

    @library_conf.setter
    def library_conf(self, conf):
        if conf is not None:
            if os.path.exists(conf):
                self._library_conf = conf
            else:
                raise ValueError('Could not find library configuration file {}'.format(conf))
        else:
            raise ValueError('Library configuration file must be specified')

    @property
    def job_conf(self):
        """Path to an electus job configuration file."""
        return self._job_conf 

    @job_conf.setter
    def job_conf(self, conf):
        if conf is not None:
            if os.path.exists(conf):
                self._job_conf = conf
            else:
                raise IOError('Could not find job configuration file {}'.format(conf))   
        else:
            raise ValueError('Job configuration file must be specified')

    @property
    def action_conf(self):
        return self._action_conf
    
    @action_conf.setter
    def action_conf(self, conf):
        if conf is not None:
            if not os.path.exists(conf):
                raise IOError('Could not find actions configuration file {}'.format(conf))
            
        self._action_conf = conf

    @property
    def evtx_file(self):
        """Path to the Sysmon event log file. Default on Windows is C:\\Windows\\System32\\winevt\\Logs\\Microsoft-Windows-Sysmon%4Operational.evtx.
        On all other systems the default is None.
        """
        return self._evtx_file

    @evtx_file.setter
    def evtx_file(self, fname):
        self._evtx_file = fname

    @property
    def etw(self):
        """Get sysmon events using Event Tracing for Windows (ETW)."""
        return self._etw

    @etw.setter
    def etw(self, e):
        if isinstance(e, bool):
            if e and platform.system() != 'Windows':
                raise NotImplementedError('Event Tracing for Windows (ETW) is only available on Windows platforms.')
            self._etw = e
        else:
            raise ValueError('ETW parameter must be True or False.')

    @property
    def evtx(self):
        """Get sysmon events from event log file."""
        return self._evtx

    @evtx.setter
    def evtx(self, e):
        if isinstance(e, bool):
            self._evtx = e
        else:
            raise ValueError('EVTX parameter must be True or False')

    @property
    def output_file(self):
        """File to log alerts to."""
        return self._output_file

    @output_file.setter
    def output_file(self, fname):
        self._output_file = fname

    @property
    def silent(self):
        """Supress printing to stdout."""
        return self._silent
    
    @silent.setter
    def silent(self, s):
        if isinstance(s, bool):
            self._silent = s
        else:
            raise ValueError('Silent must be True or False')

    @property
    def pretty_print(self):
        """Use indented JSON for all output."""
        return self._pretty_print
    
    @pretty_print.setter
    def pretty_print(self, p):
        if isinstance(p, bool):
            self._pretty_print = p
        else:
            raise ValueError('Pretty_print must be True or False')

    @property
    def continuous(self):
        """Keep reading the log when the end is reached."""
        return self._continuous
    
    @continuous.setter
    def continuous(self, c):
        if isinstance(c, bool):
            self._continuous = c
        else:
            raise ValueError("Continuous must be True or False.")

    def get_actions(self):
        """Load the actions defined in the action configuration file."""
        if self.action_conf:
            actions = []
            with open(self.action_conf, 'r') as f:
                loaded_config = json.load(f)
            for action_name in loaded_config:
                actions.append(Action(action_name, loaded_config[action_name]))

            return actions
        else:
            return []

    def take_action(self, alert):
        """Evaluate if actions should be taken and execute them if needed.
        
        :param alert: An electus Alert object.
        """

        for action in self.actions:
            if action.name == alert.name:
                for correlation in action.correlations:
                    to_match = []
                    for ind in correlation:
                        for hit in alert.hits:
                            if ind.name == hit.name:
                                ind.extract_data(hit.event)
                                to_match.append(hit.event[ind.field])
                    if (len(set(to_match)) == 1) and len(to_match) >= len(correlation):
                        for ind in correlation:
                            ind.take_action(action.name)

    def run(self):

        for event in self.sysmon.read_event():
            alerts = self.sig_engine.evaluate_event(event)

            if alerts:
                for alert in alerts:

                    if self.actions:
                        self.take_action(alert)

                    if self.output_file:
                        with open(self.output_file, 'a') as f:
                            json.dump(alert.to_dict(), f, indent=4)
                    if not self.silent:
                        if self.pretty_print:
                            print(format_pretty(alert.to_dict()))
                        else:
                            print(alert.to_dict())






import json
import xmltodict
from collections import OrderedDict
from datetime import datetime, timedelta
from dateutil import parser

def timestamp_to_epoch(timestamp_str):
    """Convert the specified timestamp to seconds since the UNIX epoch."""
    #return datetime.strptime(timestamp_str, '%Y-%m-%d %H:%M:%S.%f').timestamp()
    return parser.parse(timestamp_str).timestamp()

# Taken from http://timestamp.ooz.ie/p/time-in-python.html
def convert_ad_timestamp(timestamp):
    """Convert Windows system time format to UNIX"""
    epoch_start = datetime(year=1601, month=1,day=1)
    seconds_since_epoch = timestamp/10**7
    return epoch_start + timedelta(seconds=seconds_since_epoch)


def flatten_dict(branch, flat={}, current_path=''):
    """Flattens all the keys in the path of a nested dictionary to one string.
    
    >>> nested = {'lvl1':{'lvl2':{'lvl3': 'value'}}}
    >>> flatten_dict(nested, {})
    {'lvl1.lvl2.lvl3': 'value'}
    """
    for k, v in branch.items():
        key = k.lstrip('#@')
        path = key if not current_path else '%s.%s' % (current_path, key)
        if isinstance(v, dict):
            flatten_dict(v, flat, current_path=path)
        else:
            flat[path] = v
    return flat

def join_key_value_pairs(data_list):
    """Joins the key value pairs in the parsed event data field.
    
    :param data_list: A list of ordered dictionaries given by calling xmltodict.parse() on the EventData field.

    :returns: An OrderedDict containing the data fields and their values

    >>> example = [OrderedDict({'@Name': 'foo', '#text': 'bar'})]
    >>> join_key_value_pairs(example)
    OrderedDict([('foo', 'bar')])
    """

    flat = OrderedDict()
    for pair in data_list:
        name, value = '', ''
        try:
            name = pair['@Name']
            value = pair['#text']
        except KeyError:
            pass
        if name in flat:
            raise KeyError('Duplicate key: "%s"' % name)
        if name:
            flat[name] = value

    return flat

def flatten_xml_event(xml_event):
    """Converts an XML event into a flattened dictionary.
    
    :param xml_event: Event from Sysmon as an XML string.
    """
    xml = xmltodict.parse(xml_event)
    try:
        del xml['Event']['@xmlns']  # xml version is redundant
    except KeyError:
        pass
    
    xml['Event']['EventData']['Data'] = join_key_value_pairs(xml['Event']['EventData']['Data'])
    flat = OrderedDict()
    flatten_dict(xml['Event'], flat)
    return flat

def format_compact(record):
    """Return the record as a JSON string."""
    return json.dumps(record)

def format_pretty(record):
    """Return the record as a pretty printed JSON string"""
    return json.dumps(record, indent=4)




import argparse
import os
import json
import time
import platform
from datetime import datetime, timedelta
from queue import Queue, Empty
from threading import Thread

import Evtx.Evtx as evtx
import sys
import xmltodict
from collections import OrderedDict

from .utils import flatten_dict, timestamp_to_epoch, format_compact, format_pretty, flatten_xml_event, convert_ad_timestamp
from . import constants

if platform.system() == 'Windows':
    import etw

class ProcessNode(object):
    def __init__(self, guid, events):
        self.events = events
        self.children = {}
        self.name = ''
        self.timestamp = 0

        for event in self.events:
            if not self.name and constants.KEY_IMAGE in event:
                self.name = event[constants.KEY_IMAGE]
            if constants.KEY_TIMESTAMP in event:
                self.timestamp = min(timestamp_to_epoch(event[constants.KEY_TIMESTAMP]), self.timestamp)

        if not self.name:
            self.name = guid


class ProcessTree(object):
    def __init__(self):
        self.process_events = {}
        self.parents = {}
        self.process_chains = []
        self.tree = None

    def add_event(self, event):
        """Adds an event to the tree."""
        guid = event[constants.KEY_P_GUID] if constants.KEY_P_GUID in event else constants.GUID_ROOT
        if guid not in self.process_events:
            self.process_events[guid] = [event]
        else:
            self.process_events[guid].append(event)

        if guid and (guid not in self.parents or not self.parents[guid]):
            if constants.KEY_PP_GUID in event:
                self.parents[guid] = event[constants.KEY_PP_GUID]
            else:
                self.parents[guid] = ''

    def _build_chain(self, guid, chain):
        chain.insert(0, guid)
        if guid in self.parents:
            self._build_chain(self.parents[guid], chain)

    def _build_chains(self):
        chains = []
        for guid in self.process_events:
            chain = []
            try:
                self._build_chain(guid, chain)
            except RecursionError:
                pass
            top_chain = chain[0]
            if top_chain != '' and (top_chain not in self.process_events or not self.process_events[top_chain]):
                chain.insert(0, '')
            chains.append(chain)

        self.process_chains = sorted(chains, key=lambda x: x[0])

    def build_process_tree(self):
        """Builds an entire process tree."""
        self._build_chains()
        if constants.GUID_ROOT not in self.process_events:
            self.process_events[constants.GUID_ROOT] = []
        self.tree = ProcessNode(constants.GUID_ROOT, self.process_events[constants.GUID_ROOT])

        for chain in self.process_chains:
            node = self.tree
            for guid in chain:
                if guid not in node.children:
                    if guid in self.process_events:
                        events = self.process_events[guid]
                    else:
                        events = []
                    node.children[guid] = ProcessNode(guid, events)
                node = node.children[guid]

    @staticmethod
    def _abbreviate_event(event):
        abbreviated = OrderedDict()
        event_id = event[constants.KEY_EVENT_ID]
        event_description = constants.EVENT_KEYS[event_id][0]
        keys = constants.EVENT_KEYS_ALL + constants.EVENT_KEYS[event_id][1:]
        abbreviated['event'] = '%s (%s)' % (event_description, event_id)
        for key in keys:
            if key and key in event and event[key]:
                abbreviated[key] = event[key]

        return abbreviated

    def _abbreviate_process(self, process_node):
        abbreviated = OrderedDict()
        if process_node.events:
            events = [self._abbreviate_event(event) for event in process_node.events if event and event[constants.KEY_EVENT_ID] in constants.EVENT_KEYS]
            abbreviated_events = sorted(events, key=lambda e: timestamp_to_epoch(e[constants.KEY_TIMESTAMP]))
            if abbreviated_events:
                abbreviated['events'] = abbreviated_events

        if process_node.children:
            children_order = sorted([(child, node.timestamp) for child, node in process_node.children.items()],
                                    key=lambda c: c[1])
            abbreviated['children'] = OrderedDict()
            for child, _ in children_order:
                abbreviated['children'][process_node.children[child].name] = self._abbreviate_process(process_node.children[child])
        return abbreviated

    def abbreviate_tree(self):
        """Abbreviate the entire tree"""
        abbreviated = OrderedDict()
        if self.tree.events:
            events = [self._abbreviate_event(event) for event in self.tree.events if event]
            abbreviated['system_events'] = sorted(events, key=lambda e: timestamp_to_epoch(e[constants.KEY_TIMESTAMP]))
        abbreviated['processes'] = OrderedDict()
        for _, node in self.tree.children[''].children.items():
            abbreviated_node = self._abbreviate_process(node)
            if ('events' in abbreviated_node and abbreviated_node['events']) or \
                    ('children' in abbreviated_node and abbreviated_node['children']):
                abbreviated['processes'][node.name] = abbreviated_node
        return abbreviated


class SysmonHelper(object):

    def __init__(self, etw=False, evtx=True, evtx_file=None, continuous=False, 
                 silent=True, outfile=None, append_to_outfile=True, compact=False, 
                 very_compact=False, pretty_print=False, tree=False):

        self.evtx_queue = None
        self.etw_queue = None
        self.event_queue = None
        self.shutdown = False
        self.etw_parsers = []

        self.etw = etw
        self.evtx = evtx
        self.evtx_file = evtx_file
        self.continuous = continuous
        self.silent = silent
        self.outfile = outfile
        self.append_to_outfile = append_to_outfile
        self.compact = compact
        self.very_compact = very_compact
        self.pretty_print = pretty_print
        self.tree = tree

        if (self.compact or self.very_compact) and self.pretty_print:
            raise ValueError('Pretty_print cannot be used if compact or very_compact are set.')

        if self.very_compact and not self.compact:
            self.very_compact = True

        if self.outfile and os.path.exists(self.outfile) and not self.append_to_outfile:
            os.remove(self.outfile)

        if not self.etw and not self.evtx:
            raise ValueError("No data source specified. At least one of etw or evtx must be True.")

        if self.etw:
            # Start the ETW parser
            self.etw_queue = Queue()
            self.event_queue = self.etw_queue
            self.etw_parsers.append(self.start_etw_parser('Microsoft-Windows-Sysmon', '{5770385F-C22A-43E0-BF4C-06F5698FFBD9}'))
            self.etw_parsers.append(self.start_etw_parser('Microsoft-Windows-FileInfoMinifilter', '{A319D300-015C-48BE-ACDB-47746E154751}'))
            # etw_providers = {'Microsoft-Windows-Sysmon': '{5770385F-C22A-43E0-BF4C-06F5698FFBD9}', 
            #                 'Microsoft-Windows-FileInfoMinifilter': '{A319D300-015C-48BE-ACDB-47746E154751}'}
            # for name, guid in etw_providers.items():
            #     self.etw_parsers.append(self.start_etw_parser(name, guid))

        if self.evtx:
            # Start the EVXT parser
            self.evtx_queue = Queue()
            # Evtx should be parsed first if selected and then switch to ETW
            self.event_queue = self.evtx_queue
            self.evtx_parser = Thread(target=self.start_evtx_parser)
            self.evtx_parser.daemon = False
            self.evtx_parser.start()

    @property
    def etw(self):
        """Get sysmon events using Event Tracing for Windows (ETW)."""
        return self._etw

    @etw.setter
    def etw(self, e):
        if isinstance(e, bool):
            if e and platform.system() != 'Windows':
                raise NotImplementedError('Event Tracing for Windows (ETW) is only available on Windows platforms.')
            self._etw = e
        else:
            raise ValueError('ETW parameter must be True or False.')

    @property
    def evtx(self):
        """Get sysmon events from event log file."""
        return self._evtx

    @evtx.setter
    def evtx(self, e):
        if isinstance(e, bool):
            self._evtx = e
        else:
            raise ValueError('EVTX parameter must be True or False')

    @property
    def evtx_file(self):
        """Path to the SYSMON ".evtx" file. Default on Windows is C:\\Windows\\System32\\winevt\\Logs\\Microsoft-Windows-Sysmon%4Operational.evtx.
        On all other systems the default is None.
        """
        return self._evtx_file

    @evtx_file.setter
    def evtx_file(self, file_name):
        if file_name is not None:
            if os.path.exists(file_name):
                self._evtx_file = file_name
            else:
                raise ValueError('Cannot find evtx_file {}.'.format(file_name))
        elif file_name is None and platform.system() == 'Windows' and self.evtx:
            default = r"C:\Windows\System32\winevt\Logs\Microsoft-Windows-Sysmon%4Operational.evtx"
            if os.path.exists(default):
                self._evtx_file = default
            else:
                raise ValueError('Cannot find default evtx_file {}. Is sysmon installed and running?'.format(default))
        elif file_name is None and self.evtx:
            raise ValueError('Evtx flag was set but no evtx_file was given')
        else:
            self._evtx_file = None

    @property
    def continuous(self):
        """Keep reading the log when the end is reached."""
        return self._continuous
    
    @continuous.setter
    def continuous(self, c):
        if isinstance(c, bool):
            self._continuous = c
        else:
            raise ValueError("Continuous must be True or False.")
        
    @property
    def silent(self):
        """Supress printing to stdout."""
        return self._silent
    
    @silent.setter
    def silent(self, s):
        if isinstance(s, bool):
            self._silent = s
        else:
            raise ValueError('Silent must be True or False')

    @property
    def outfile(self):
        """Write output to this file."""
        return self._outfile

    @outfile.setter
    def outfile(self, o):
        self._outfile = o

    @property
    def append_to_outfile(self):
        """Append to output file instead of overwriting."""
        return self._append_to_outfile

    @append_to_outfile.setter
    def append_to_outfile(self, a):
        if isinstance(a, bool):
            self._append_to_outfile = a
        else:
            raise ValueError('Append_to_outfile must be True or False.')

    @property
    def compact(self):
        """Use one JSON record per line for all output."""
        return self._compact

    @compact.setter
    def compact(self, c):
        if isinstance(c, bool):
            self._compact = c
        else:
            raise ValueError('Compact must be True or False.')

    @property
    def very_compact(self):
        """Print the first 80 characters of each record."""
        return self._very_compact
    
    @very_compact.setter
    def very_compact(self, v):
        if isinstance(v, bool):
            self._very_compact = v
        else:
            raise ValueError('Very_compact must be True or False.')

    @property
    def tree(self):
        """Write a process tree to the output-file."""
        return self._tree
    
    @tree.setter
    def tree(self, t):
        if isinstance(t, bool):
            self._tree = t
        else:
            raise ValueError('Tree must be True or False.')

    @property
    def pretty_print(self):
        """Use indented JSON for all output."""
        return self._pretty_print
    
    @pretty_print.setter
    def pretty_print(self, p):
        if isinstance(p, bool):
            self._pretty_print = p
        else:
            raise ValueError('Pretty_print must be True or False')

    def start_etw_parser(self, name, guid):
        providers = [etw.ProviderInfo(name, etw.GUID(guid))]
        job = etw.ETW(session_name=name, event_callback=self.on_etw_event, providers=providers)
        job.start()
        
        return job

    def start_evtx_parser(self):
        for event in self.read_sysmon_log(self.evtx_file, continuous=self.continuous):
            if self.evtx_queue:
                self.evtx_queue.put((event,"evtx"))
            else:
                break

    def read_sysmon_log(self, sysmon_file, continuous=False):
        log = evtx.Evtx(sysmon_file)
        log = log.__enter__()

        # TODO: Add an argument so a specific record ID can be specified (for testing)
        # Finds the first record ID (required if the log has been cleared)
        counter = log.get_file_header().first_chunk().log_first_record_number()
        while not self.shutdown:
            event_record = log.get_record(counter)
            if event_record is None:
                if continuous:
                    time.sleep(0.2)
                    continue
                else:
                    break
            else:
                counter += 1
            event = flatten_xml_event(event_record.xml())
            if not event:
                continue
            yield event

    def read_event(self):
        changeover_event = {}
        while not self.shutdown:
            try:
                event, source = self.event_queue.get(block=True, timeout=3)
                if source is "evtx":
                    if changeover_event:
                        if event['EventData.Data.UtcTime'] == changeover_event['EventData.Data.UtcTime'] \
                                and event['EventData.Data.ProcessGuid'] == changeover_event['EventData.Data.ProcessGuid'] \
                                and event['System.Execution.ProcessID'] == changeover_event['System.Execution.ProcessID'] \
                                and event['System.Execution.ThreadID'] == changeover_event['System.Execution.ThreadID']:
                            self.event_queue = self.etw_queue
                            self.evtx_queue = None
                    else:
                        if self.etw and self.etw_queue.not_empty:
                            changeover_event = self.etw_queue.get(block=False)
                yield event
            except Empty:
                try:
                    if self.continuous:
                        sleep_len = 0.2
                        print('Sleeping for {} seconds'.format(sleep_len))
                        time.sleep(sleep_len)
                    else:
                        self.cleanup()
                        break
                except KeyboardInterrupt:
                    self.cleanup()
            except KeyboardInterrupt:
                self.cleanup()

    def cleanup(self):
        self.shutdown = True
        for parser in self.etw_parsers:
            parser.stop()

    def on_etw_event(self, event):
        formatted_event = {}
        for field in event[1]:
            if field is "EventHeader":
                # Manually rename event header fields as they don't match the EVTX field names
                formatted_event["System.Provider.Name"] = "Microsoft-Windows-Sysmon"
                formatted_event["System.Provider.Guid"] = event[1][field]['ProviderId']
                formatted_event["System.EventID.Qualifiers"] = "NA"
                formatted_event["System.EventID.text"] = str(event[1][field]['EventDescriptor']['Id'])
                formatted_event["System.Version"] = str(event[1][field]['EventDescriptor']['Version'])
                formatted_event["System.Level"] = str(event[1][field]['EventDescriptor']['Level'])
                formatted_event["System.Task"] = str(event[1][field]['EventDescriptor']['Task'])
                formatted_event["System.Opcode"] = str(event[1][field]['EventDescriptor']['Opcode'])
                formatted_event["System.Keywords"] = str(hex(event[1][field]['EventDescriptor']['Keyword']))
                formatted_event["System.TimeCreated.SystemTime"] = str(convert_ad_timestamp(int(event[1][field]['TimeStamp'])))
                formatted_event["System.EventRecordID"] = "NA"
                formatted_event["System.Correlation.ActivityID"] = str(event[1][field]['ActivityId'])
                formatted_event["System.Correlation.RelatedActivityID"] = "NA"
                formatted_event["System.Execution.ProcessID"] = str(event[1][field]['ProcessId'])
                formatted_event["System.Execution.ThreadID"] = str(event[1][field]['ThreadId'])
                formatted_event["System.Channel"] = "NA"
                formatted_event["System.Computer"] = "NA"
                formatted_event["System.Security.UserID"] = "NA"
            else:
                formatted_event["EventData.Data." + field] = str(event[1][field])
        # Add event to the queue
        self.etw_queue.put((OrderedDict(formatted_event),"etw"))

    def event_processor(self):
        process_tree = ProcessTree()
        try:
            for event in self.read_event():
                if self.tree:
                    process_tree.add_event(event)

                if not self.silent:
                    output_string = format_compact(event) if self.compact else format_pretty(event)
                    output_string = "{} - PID: {} TID: {} GUID: {}".format(event['EventData.Data.UtcTime'], event['System.Execution.ProcessID'], event['System.Execution.ThreadID'], event['EventData.Data.ProcessGuid']) if self.very_compact else output_string
                    print(output_string)

                if self.outfile and not self.tree:
                    output_string = format_pretty(event) if self.pretty_print else format_compact(event)
                    output_string = "{} - PID: {} TID: {} GUID: {}".format(event['EventData.Data.UtcTime'], event['System.Execution.ProcessID'], event['System.Execution.ThreadID'], event['EventData.Data.ProcessGuid']) if self.very_compact else output_string
                    try:
                        with open(self.outfile, 'a') as f:
                            f.write('%s\n' % output_string)
                    except IOError:
                        print('Error writing to %s' % self.outfile)
        except KeyboardInterrupt:
            self.cleanup()
        if self.tree:
            process_tree.build_process_tree()
            tree = process_tree.abbreviate_tree()
            mode = 'a' if self.append_to_outfile else 'w'
            with open(self.outfile, mode) as f:
                json.dump(tree, f, indent=4)



KEY_EVENT_ID = 'System.EventID.text'
KEY_P_GUID = 'EventData.Data.ProcessGuid'
KEY_PP_GUID = 'EventData.Data.ParentProcessGuid'
KEY_TIMESTAMP = 'System.TimeCreated.SystemTime'
KEY_IMAGE = 'EventData.Data.Image'
GUID_ROOT = ''

EVENT_KEYS_ALL = [
    KEY_TIMESTAMP,
    # KEY_P_GUID,
]

EVENT_KEYS = {
    '1': [
        'Process Create',
        'EventData.Data.Image',
        'EventData.Data.CommandLine',
        'EventData.Data.ProcessGuid',
        'EventData.Data.ProcessId',
        'EventData.Data.LogonGuid',
        'EventData.Data.ParentImage',
        'EventData.Data.ParentProcessGuid',
        'EventData.Data.ParentProcessId',
        'EventData.Data.Hashes',
    ],
    '2': [
        'Changed File Creation Time',
        'EventData.Data.Image',
    ],
    '3': [
        'Network Connection',
        'EventData.Data.Protocol',
        'EventData.Data.SourceIp',
        'EventData.Data.SourcePort',
        'EventData.Data.DestinationIp',
        'EventData.Data.DestinationPort',
    ],
    '4': [
        'SYSMON State Changed',
    ],
    '5': [
        'Process Terminated',
    ],
    '6': [
        'Driver Load',
        'EventData.Data.ImageLoaded',
        'Realtek Semiconductor Corp.',
    ],
    '7': [
        'Image Load',
        'EventData.Data.ImageLoaded',
        'EventData.Data.Signature',
        'EventData.Data.SignatureStatus',
    ],
    '8': [
        'Create Remote Thread',
        'EventData.Data.SourceImage',
        'EventData.Data.SourceProcessGuid',
        'EventData.Data.SourceProcessId',
        'EventData.Data.TargetImage',
        'EventData.Data.TargetProcessGuid',
        'EventData.Data.TargetProcessId',
        'EventData.Data.StartAddress',
        'EventData.Data.StartModule',
        'EventData.Data.StartFunction',
    ],
    '9': [
        'Raw Access Read',
    ],
    '10': [
        'Process Access',
    ],
    '11': [
        'File Create',
        'EventData.Data.TargetFilename',
    ],
    '12': [
        'Registry Object',
        'EventData.Data.EventType',
        'EventData.Data.TargetObject',
    ],
    '13': [
        'Registry Value Set',
        'EventData.Data.TargetObject',
        'EventData.Data.Details',
    ],
    '14': [
        'Registry Object Renamed',
        'EventData.Data.EventType',
        'EventData.Data.TargetObject',
        'EventData.Data.NewName',
    ],
    '15': [
        'File Create Stream Hash',
        'EventData.Data.TargetFilename',
        'EventData.Data.Hashes',
    ],
    '16': [
        'SYSMON Config Change',
    ],
    '17': [
        'Pipe Created',
    ],
    '18': [
        'Pipe Connected',
    ],
    '19': [
        'WMI Event Filter',
    ],
    '20': [
        'WMI Event Consumer',
    ],
    '21': [
        'WMI Event Connector',
    ],
    '255': [
        'SYSMON ERROR',
    ],
}


import platform
from setuptools import setup

# Base dependencies for all platforms
install_requires = ['python-evtx', 'xmltodict', 'electus']

if platform.system() == 'Windows':
    install_requires.append(['psutil', 'easygui', 'pywintrace'])

setup(name='phishermon',
      version='0.1.13',
      description='A tool to create and take action on behavioural signatures from sysmon data.',
      url='https://gitlab.com/gclenden/phishermon',
      author='Graham Clendenning',
      author_email='graham.clendenning@uoit.net',
      license='MIT',
      packages=['phishermon'],
      entry_points={'console_scripts':['phishermon=phishermon.commandline:phishermon_cmdline', 
                                       'sysmon-parser=phishermon.commandline:sysmon_parser_cmdline']},
      install_requires = install_requires,
      classifiers = ["Natural Language :: English",
            'Development Status :: 4 - Beta',
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
            "Programming Language :: Python",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.4",
            "Programming Language :: Python :: 3.5",
            "Programming Language :: Python :: 3.6"],
      zip_safe=False)


import os
from .mitigations import MitigateFile, MitigateProcess, MitigateRegistry

mitigate_proc = MitigateProcess()
mitigate_file = MitigateFile()
mitigate_reg = MitigateRegistry()

class Action(object):

    def __init__(self, name, correlations):

        self.name = name
        self.correlations = correlations

    @property
    def name(self):
        """A unique name for the action."""
        return self._name

    @name.setter
    def name(self, n):
        if isinstance(n, str):
            self._name = n
        else:
            raise ValueError('Name must be a string')

    @property
    def correlations(self):
        """List of :class:`~phishermon.action.CorrelationIndicator` objects. Defined in the action configuration file."""
        return self._correlations

    @correlations.setter
    def correlations(self, corrs):
        if not isinstance(corrs, list):
            raise ValueError('Value for action {} must be a list'.format(self.name))

        self._correlations = [CorrelationIndicator(c) for c in corrs]


class CorrelationIndicator(object):

    valid_types = {'process': ['process', 'pid'], 'file': ['path'], 'registry': ['path', 'key']}

    def __init__(self, conf):

        self.name = conf['name']
        self.field = conf['field']
        self.event_type = conf.get('type', None)
        self.extract = conf.get('to_extract', {})

        if self.event_type and not self.extract:
            raise ValueError('Type is specified but there are no fields to extract. Please specify a to_extract field.')

    @property
    def name(self):
        """Name of the job in the job configuration file"""
        return self._name

    @name.setter
    def name(self, n):
        if isinstance(n, str):
            self._name = n
        else:
            raise ValueError('Name must be a string')

    @property
    def field(self):
        """Field name to compare between hit objects. An action will only be taken if all the values of field in the hit objects match."""
        return self._field

    @field.setter
    def field(self, f):
        if isinstance(f, str):
            self._field = f
        else:
            raise ValueError('Field must be a string')

    @property
    def extract(self):
        """A dictionary of fields to extract to take action.
        The required fields for each type are {'process': ['process', 'pid'], 'file': ['path'], 'registry': ['path', 'key']}
        """
        return self._extract

    @extract.setter
    def extract(self, e):

        if e and self.event_type is None:
            raise ValueError('extract is set but type is not specified')

        if not e:
            self._extract = None
        elif isinstance(e, dict):
            e = {k.lower(): v for k,v in e.items()}
            for f in self.valid_types[self.event_type]:
                if f not in e:
                    raise ValueError('Type {} must have keys {} in the extract field. Given keys were {}'
                                        .format(self.event_type, self.valid_types[self.event_type], list(e.keys())))

            self._extract = e
        else:
            raise ValueError('to_extract must be a dictionary')

    @property
    def event_type(self):
        """The type of event to take action on. Valid event types are process, file, or registry."""
        return self._event_type

    @event_type.setter
    def event_type(self, t):
        if t is None:
            self._event_type = None
        elif not isinstance(t, str):
            raise ValueError('type must be a string')
        elif t.lower() not in self.valid_types:
            raise ValueError('type {} is not a valid type. Valid types are: {}'.format(t, self.valid_types.keys()))
        else:
            self._event_type = t.lower()

    def extract_data(self, event):
        """Get the data from the fields defined by to_extract."""
        if self.extract:
            self.extracted_data = {k: event[v] for k, v in self.extract.items()}
        else:
            self.extracted_data = None

    def take_action(self, action_name=None):
        """Take the action specified."""
        if not self.extracted_data:
            return None

        if self.event_type == 'process':
            proc = self.extracted_data.get('process', None)
            if proc is not None:
                proc = os.path.basename(proc)
            pid = self.extracted_data.get('pid', None)

            if proc is not None and pid is not None:
                mitigate_proc.kill_process(int(pid), proc)

        elif self.event_type == 'file':
            path = self.extracted_data.get('path', None)

            if path is not None:
                mitigate_file.delete_file(path)
        
        elif self.event_type == 'registry':
            path = self.extracted_data.get('path', None)
            key = self.extracted_data.get('key', None)

            if path is not None and key is not None:
                mitigate_reg.delete_registry_key(path, key)
















import argparse
from phishermon import Phishermon, SysmonHelper

def phishermon_cmdline():

    parser = argparse.ArgumentParser('Read sysmon data and apply behavioural signatures.')
    parser.add_argument('--library', '-l', help='Path to library configuration file', type=str, default=None)
    parser.add_argument('--jobs', '-j', help='Path to job configuration file', type=str, default=None)
    parser.add_argument('--actions', '-a' , help='Path to actions configuration file', type=str, default=None)
    parser.add_argument('--etw', '-e', action='store_true', help='Get events using Event Tracing for Windows', default=False)
    parser.add_argument('--evtx', '-E', action='store_true', help='Use saved sysmon log', default=False)
    parser.add_argument('--evtx-file', '-f', dest='evtx_file', help='Path to the sysmon .evtx file', default=None)
    parser.add_argument('--silent', '-s',  action='store_true', help='Suppress standard output', default=False)
    parser.add_argument('--output', '-o', help='Path to output file', type=str, default=None)
    parser.add_argument('--pretty', '-p',  action='store_true', help='Use indented JSON for all output.', default=False)
    parser.add_argument('--continuous', action='store_true', help='Stop reading the log when the end is reached.', default=False)
    
    args = parser.parse_args()

    p = Phishermon(library_conf=args.library, job_conf=args.jobs, action_conf=args.actions, evtx_file=args.evtx_file, 
                   etw=args.etw, evtx=args.evtx, output_file=args.output, silent=args.silent, pretty_print=args.pretty, continuous=args.continuous)

    p.run()

def sysmon_parser_cmdline():
    parser = argparse.ArgumentParser('Read and output sysmon logs in a reader-friendly format. Use CTRL+C to stop monitoring.')
    parser.add_argument('--etw', '-e', action='store_true', help='Get events using Event Tracing for Windows', default=False)
    parser.add_argument('--evtx', '-E', action='store_true', help='Use saved sysmon log', default=False)
    parser.add_argument('--evtx-file', '-f', dest='evtx_file', help='Path to the sysmon .evtx file.', default=None)
    parser.add_argument('--silent', '-s', action='store_true', help='Suppress standard output.', default=False)
    parser.add_argument('--output', '-o', help='Also write output to this file.', type=str, default=None)
    parser.add_argument('--compact', '-c', action='store_true', help='Use one JSON record per line for all output.', default=False)
    parser.add_argument('--verycompact', '-C', dest='very_compact', action='store_true', help='Print the first 80 characters of each record', default=False)
    parser.add_argument('--pretty', '-p',  action='store_true', help='Use indented JSON for all output.', default=False)
    parser.add_argument('--append', '-a', action='store_true', help='Append to output file instead of overwriting.', default=False)
    parser.add_argument('--continuous', action='store_true', help='Stop reading the log when the end is reached.', default=False)
    parser.add_argument('--tree', '-t', action='store_true', help='Write a process tree to the output-file.', default=False)

    args = parser.parse_args()

    sysmon = SysmonHelper(etw=args.etw, evtx=args.evtx, evtx_file=args.evtx_file, continuous=args.continuous, silent=args.silent, 
                          outfile=args.output, append_to_outfile=args.append, compact=args.compact, very_compact=args.very_compact, 
                          pretty_print=args.pretty, tree=args.tree)

    sysmon.event_processor()
